/*
 This code computes the Dyanmic Time Warping (DTW) or the Local Frame
 Match Distance (LFMD) between two temporal paths (hand gestures).
 
 Copyright (C) 2017  Radu Tudor Ionescu, Marius Popescu,
                     Christopher Conly, Vassilis Athitsos
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

public class DistanceAlgorithms
{
    public DistanceAlgorithms()
    {
    }

    public static double matchCost(double[] features1, double[] features2)
    {
        int i,j;
        
        double norm1, norm2;
        double cost = 0;
        
        cost += Math.sqrt(Math.pow(features1[0] - features2[0], 2) + Math.pow(features1[1] - features2[1], 2));
        
        return cost;
    }
    
    public static double DTW(double[][] sign1, double[][] sign2)
    {
        int i, j;
        
        double[][] scores;
        
        scores = new double[sign1.length][sign2.length];
        
        scores[0][0] = matchCost(sign1[0], sign2[0]);
        
        for (i = 1; i < sign1.length; i++)
        {
            scores[i][0] = scores[i-1][0] + matchCost(sign1[i], sign2[0]);
        }
        
        for (j = 1; j < sign2.length; j++)
        {
            scores[0][j] = scores[0][j-1] + matchCost(sign1[0], sign2[j]);
        }
        
        for (i = 1; i < sign1.length; i++)
        {
            for (j = 1; j < sign2.length; j++)
            {
                scores[i][j] =  matchCost(sign1[i], sign2[j]) + Math.min(scores[i-1][j], Math.min(scores[i][j-1], scores[i-1][j-1]));
            }
        }
        
        double d = scores[sign1.length-1][sign2.length-1];
        
        return d;
    }
    
    /* The parameter th controls the trade-off between accuracy and speed.
       (smaller => faster, bigger => more accurate) */
    public static double LFMD(double[][] sign1, double[][] sign2, int th)
    {
        int i, j;
        
        double[][] scores;
        double[] min1;
        double[] min2;
        int[] idxMin1;
        int[] idxMin2;
        
        scores = new double[sign1.length][sign2.length];
        min1 = new double[sign1.length];
        min2 = new double[sign2.length];
        idxMin1 = new int[sign1.length];
        idxMin2 = new int[sign2.length];
        
        for (i = 0; i < sign1.length; i++)
        {
            min1[i] = Double.POSITIVE_INFINITY;
            idxMin1[i] = 0;
        }
        
        for (j = 0; j < sign2.length; j++)
        {
            min2[j] = Double.POSITIVE_INFINITY;
            idxMin2[j] = 0;
        }
        
        for (i = 0; i < sign1.length; i++)
        {
            for (j = 0; j < sign2.length; j++)
            {
                if (Math.abs(i - j) < th) // ***
                {
                    scores[i][j] = matchCost(sign1[i], sign2[j]);
                    
                    if (scores[i][j] < min1[i])
                    {
                        min1[i] = scores[i][j];
                        idxMin1[i] = j;
                    }
                    
                    if (scores[i][j] < min2[j])
                    {
                        min2[j] = scores[i][j];
                        idxMin2[j] = i;
                    }
                }
            }
        }
        
        double d = 0.0;

        for (i = 1; i < sign1.length; i++)
        {
            d += min1[i] * (Math.abs(idxMin1[i] - i) + 1);
        }
        
        for (j = 0; j < sign2.length; j++)
        {
            d += min2[j] * (Math.abs(idxMin2[j] - j) + 1);
        }
        
        return d / (sign1.length * sign2.length);
    }
    
    /* The parameter th controls the trade-off between accuracy and speed.
     (smaller => faster, bigger => more accurate) */
    public static double LFMD2(double[][] sign1, double[][] sign2, int th)
    {
        int i, j;
        
        double[][] scores;
        double[] min1;
        double[] min2;
        int[] idxMin1;
        int[] idxMin2;
        
        scores = new double[sign1.length][sign2.length];
        min1 = new double[sign1.length];
        min2 = new double[sign2.length];
        idxMin1 = new int[sign1.length];
        idxMin2 = new int[sign2.length];
        
        for (i = 0; i < sign1.length; i++)
        {
            scores[i][0] = matchCost(sign1[i], sign2[0]);
            min1[i] = Double.POSITIVE_INFINITY;
            idxMin1[i] = -1;
        }
        
        for (j = 0; j < sign2.length; j++)
        {
            scores[0][j] = matchCost(sign1[0], sign2[j]);
            min2[j] = Double.POSITIVE_INFINITY;
            idxMin2[j] = -1;
        }
        
        for (i = 0; i < sign1.length - 1; i++)
        {
            for (j = 0; j < sign2.length - 1; j++)
            {
                if (Math.abs(i - j) < th) // ***
                {
                    scores[i+1][j+1] = matchCost(sign1[i+1], sign2[j+1]);
                    scores[i][j] += scores[i+1][j+1];
                    
                    if (scores[i][j] < min1[i])
                    {
                        min1[i] = scores[i][j];
                        idxMin1[i] = j;
                    }
                    
                    if (scores[i][j] < min2[j])
                    {
                        min2[j] = scores[i][j];
                        idxMin2[j] = i;
                    }
                }
            }
        }
        
        double d = 0.0;
        
        for (i = 1; i < sign1.length - 1; i++)
        {
            d += min1[i] * (Math.abs(idxMin1[i] - i) + 1);
        }
        
        for (j = 0; j < sign2.length - 1; j++)
        {
            d += min2[j] * (Math.abs(idxMin2[j] - j) + 1);
        }
        
        return d / (sign1.length * sign2.length);
    }
    
    /* The parameter th controls the trade-off between accuracy and speed.
     (smaller => faster, bigger => more accurate) */
    public static double LFMD3(double[][] sign1, double[][] sign2, int th)
    {
        int i, j;

        double[][] scores;
        double[] min1;
        double[] min2;
        int[] idxMin1;
        int[] idxMin2;
        
        scores = new double[sign1.length][sign2.length];
        min1 = new double[sign1.length];
        min2 = new double[sign2.length];
        idxMin1 = new int[sign1.length];
        idxMin2 = new int[sign2.length];
        
        for (i = 0; i < sign1.length - 1; i++)
        {
            scores[i+1][1] = matchCost(sign1[i+1], sign2[1]);
            scores[i][0] = matchCost(sign1[i], sign2[0]) + scores[i+1][1];

            min1[i] = Double.POSITIVE_INFINITY;
            idxMin1[i] = -1;
        }
        
        for (j = 0; j < sign2.length - 1; j++)
        {
            scores[1][j+1] = matchCost(sign1[1], sign2[j+1]);
            scores[0][j] = matchCost(sign1[0], sign2[j]) + scores[1][j+1];

            min2[j] = Double.POSITIVE_INFINITY;
            idxMin2[j] = -1;
        }
        
        for (i = 0; i < sign1.length - 2; i++)
        {
            for (j = 0; j < sign2.length - 2; j++)
            {
                if (Math.abs(i - j) < th) // ***
                {
                    scores[i+2][j+2] = matchCost(sign1[i+2], sign2[j+2]);
                    scores[i+1][j+1] += scores[i+2][j+2];
                    scores[i][j] += scores[i+2][j+2];
                    
                    if (scores[i][j] < min1[i])
                    {
                        min1[i] = scores[i][j];
                        idxMin1[i] = j;
                    }
                    
                    if (scores[i][j] < min2[j])
                    {
                        min2[j] = scores[i][j];
                        idxMin2[j] = i;
                    }
                }
            }
        }
        
        double d = 0.0;
        
        for (i = 1; i < sign1.length - 2; i++)
        {
            d += min1[i] * (Math.abs(idxMin1[i] - i) + 1);
        }
        
        for (j = 0; j < sign2.length - 2; j++)
        {
            d += min2[j] * (Math.abs(idxMin2[j] - j) + 1);
        }
        
        return d / (sign1.length * sign2.length);
    }
    
    /* The parameter th controls the trade-off between accuracy and speed.
     (smaller => faster, bigger => more accurate) */
    public static double LFMD4(double[][] sign1, double[][] sign2, int th)
    {
        int i, j;

        double[][] scores;
        double[] min1;
        double[] min2;
        int[] idxMin1;
        int[] idxMin2;
        
        scores = new double[sign1.length][sign2.length];
        min1 = new double[sign1.length];
        min2 = new double[sign2.length];
        idxMin1 = new int[sign1.length];
        idxMin2 = new int[sign2.length];
        
        for (i = 0; i < sign1.length - 2; i++)
        {
            scores[i+2][2] = matchCost(sign1[i+2], sign2[2]);
            scores[i+1][1] = matchCost(sign1[i+1], sign2[1]) + scores[i+2][2];
            scores[i][0] = matchCost(sign1[i], sign2[0]) + scores[i+1][1];
            
            min1[i] = Double.POSITIVE_INFINITY;
            idxMin1[i] = -1;
        }
        
        for (j = 0; j < sign2.length - 2; j++)
        {
            scores[2][j+2] = matchCost(sign1[2], sign2[j+2]);
            scores[1][j+1] = matchCost(sign1[1], sign2[j+1]) + scores[2][j+2];
            scores[0][j] = matchCost(sign1[0], sign2[j]) + scores[1][j+1];
            
            min2[j] = Double.POSITIVE_INFINITY;
            idxMin2[j] = -1;
        }
        
        for (i = 0; i < sign1.length - 3; i++)
        {
            for (j = 0; j < sign2.length - 3; j++)
            {
                if (Math.abs(i - j) < th) // ***
                {
                    scores[i+3][j+3] = matchCost(sign1[i+3], sign2[j+3]);
                    scores[i+2][j+2] += scores[i+3][j+3];
                    scores[i+1][j+1] += scores[i+3][j+3];
                    scores[i][j] += scores[i+3][j+3];
                    
                    if (scores[i][j] < min1[i])
                    {
                        min1[i] = scores[i][j];
                        idxMin1[i] = j;
                    }
                    
                    if (scores[i][j] < min2[j])
                    {
                        min2[j] = scores[i][j];
                        idxMin2[j] = i;
                    }
                }
            }
        }
        
        double d = 0.0;
        
        for (i = 1; i < sign1.length - 3; i++)
        {
            d += min1[i] * (Math.abs(idxMin1[i] - i) + 1);
        }
        
        for (j = 0; j < sign2.length - 3; j++)
        {
            d += min2[j] * (Math.abs(idxMin2[j] - j) + 1);
        }
        
        return d / (sign1.length * sign2.length);
    }
}
